const gfx = @import("gfx");

const ona = @import("ona");

const std = @import("std");

pub const Box = struct {
	left: f32,
	top: f32,
	width: f32,
	height: f32,
};

pub const Canvas = struct {
	nodes: ona.stack.Parallel(Node) = .{},
	drawable_blocks: ona.stack.Sequential(DrawableBlock) = .{},
	drawable_labels: ona.stack.Sequential(DrawableLabel) = .{},

	fn Drawable(comptime State: type) type {
		return struct {
			state: State,
			depth: f32,
		};
	}

	const DrawableBlock = Drawable(struct {
		background: gfx.Texture = .default,
		color: gfx.Color = gfx.colors.white,
	});

	const DrawableLabel = Drawable(struct {
		text: ona.stack.Sequential(u8) = .{},
		color: gfx.Color = gfx.colors.black,
		font: gfx.Font = .default,
	});

	const Node = struct {
		name: []u8,
		box: Box,
		parent_index: u32 = unused_index,
		block_index: u32 = unused_index,
		label_index: u32 = unused_index,
		style_index: u32 = unused_index,

		const unused_index = std.math.maxInt(u32);
	};

	pub fn append(self: *Canvas, name: []const u8, box: Box) std.mem.Allocator.Error!Item {
		const duped_name = try ona.heap.allocator.dupe(u8, name);

		errdefer {
			ona.heap.allocator.free(duped_name);
		}

		const index = self.nodes.len();

		try self.nodes.push_grow(.{
			.name = duped_name,
			.box = box,
		});

		errdefer {
			std.debug.assert(self.nodes.pop());
		}

		const node_count = self.nodes.len();

		try self.drawable_blocks.grow(ona.scalars.sub(node_count, self.drawable_blocks.cap) orelse 0);
		try self.drawable_labels.grow(ona.scalars.sub(node_count, self.drawable_labels.cap) orelse 0);

		return @enumFromInt(index);
	}

	fn calculate_depth(self: Canvas, item: Item) f32 {
		const parent_indices = self.nodes.values.slice(.parent_index);
		var depth: f32 = 0;
		var index = @intFromEnum(item);

		while (true) : (depth += 1) {
			const parent_index = parent_indices[index];

			if (parent_index == Node.unused_index) {
				break;
			}

			index = parent_index;
		}

		return depth;
	}

	fn deinit(self: *Canvas) void {
		self.nodes.deinit();
		self.drawable_blocks.deinit();

		for (self.drawable_labels.values) |*drawable_label| {
			drawable_label.state.text.deinit();
		}

		self.drawable_labels.deinit();

		self.* = undefined;
	}

	const label_depth_offset = 0.5;

	pub fn set_block(self: *Canvas, item: Item, set: SetBlock) void {
		const block_index = self.nodes.values.get(.block_index, @intFromEnum(item)).?;

		if (block_index.* == Node.unused_index) {
			block_index.* = @intCast(self.drawable_blocks.len());

			std.debug.assert(self.drawable_blocks.push(.{
				.depth = self.calculate_depth(item),
				.state = .{},
			}));
		}

		const block = &self.drawable_blocks.values[block_index.*];

		if (set.has_background) |background| {
			block.state.background = background;
		}

		if (set.has_color) |color| {
			block.state.color = color;
		}
	}

	pub fn set_label(self: *Canvas, item: Item, set: SetLabel) std.mem.Allocator.Error!void {
		const label_index = self.nodes.values.get(.label_index, @intFromEnum(item)).?;

		if (label_index.* == Node.unused_index) {
			label_index.* = @intCast(self.drawable_labels.len());

			std.debug.assert(self.drawable_labels.push(.{
				.depth = self.calculate_depth(item) + label_depth_offset,
				.state = .{},
			}));
		}

		const label = &self.drawable_labels.values[label_index.*];

		if (set.has_text) |text| {
			label.state.text.pop_all();

			try label.state.text.grow(ona.scalars.sub(text.len, label.state.text.cap) orelse 0);

			std.debug.assert(label.state.text.push_all(text));
		}

		if (set.has_font) |font| {
			label.state.font = font;
		}

		if (set.has_color) |color| {
			label.state.color = color;
		}
	}

	pub fn set_parent(self: *Canvas, item: Item, has_parent: ?Item) void {
		const item_index = @intFromEnum(item);
		const parent_index = self.nodes.values.get(.parent_index, item_index).?;

		parent_index.* = if (has_parent) |parent| @intFromEnum(parent) else Node.unused_index;

		const label_index = self.nodes.values.get(.parent_index, item_index).?.*;
		const block_index = self.nodes.values.get(.parent_index, item_index).?.*;
		const label_index_exists = label_index != Node.unused_index;
		const block_index_exists = block_index != Node.unused_index;

		if (label_index_exists or block_index_exists) {
			const depth = self.calculate_depth(item);

			if (label_index_exists) {
				self.drawable_labels[label_index].depth = depth + label_depth_offset;
			}

			if (block_index_exists) {
				self.drawable_labels[label_index].depth = depth;
			}
		}
	}
};

pub const Item = enum (u32) {
	_,
};

pub const SetBlock = struct {
	has_background: ?gfx.Texture = null,
	has_color: ?gfx.Color = null,
};

pub const SetLabel = struct {
	has_text: ?[]const u8 = null,
	has_color: ?gfx.Color = null,
	has_font: ?gfx.Font = null,
};

pub fn exit(canvas: ona.Write(Canvas)) !void {
	canvas.state.deinit();
}

pub fn render(commands: gfx.Commands(.overlay), canvas: ona.Read(Canvas)) !void {
	const boxes = canvas.state.nodes.values.slice(.box);

	for (canvas.state.drawable_blocks.values, boxes) |block, box| {
		try commands.draw_texture(.{
			.depth = block.depth,
			.position = .{box.left, box.top},
			.size = .{box.width, box.height},
			.texture = block.state.background,
			.tint = block.state.color,
		});
	}

	for (canvas.state.drawable_labels.values, boxes) |label, box| {
		try commands.draw_font(.{
			.depth = label.depth,
			.text = label.state.text.values,
			.position = .{box.left, box.top},
			.size = .{box.width, box.height},
			.tint = label.state.color,
			.font = label.state.font,
		});
	}
}

pub fn setup(world: *ona.World) !void {
	try world.set_state(Canvas{});
	try world.on_event(.render, ona.system_fn(render), .{.label = "render gui"});
	try world.on_event(.exit, ona.system_fn(exit), .{.label = "cleanup gui"});
}
